@@ -23,14 +23,17 @@
  */
 
 #include <rfb/rfb.h>
+#include <rfb/rfbregion.h>
+#include "private.h"
+
+void rfbScaledScreenUpdate(rfbScreenInfoPtr screen, int x1, int y1, int x2, int y2);
 
 /*
  * Send cursor shape either in X-style format or in client pixel format.
  */
 
 rfbBool
-rfbSendCursorShape(cl)
-    rfbClientPtr cl;
+rfbSendCursorShape(rfbClientPtr cl)
 {
     rfbCursorPtr pCursor;
     rfbFramebufferUpdateRectHeader rect;
@@ -41,6 +44,8 @@ rfbSendCursorShape(cl)
     uint8_t *bitmapData;
     uint8_t bitmapByte;
 
+    /* TODO: scale the cursor data to the correct size */
+
     pCursor = cl->screen->getCursorPtr(cl);
     /*if(!pCursor) return TRUE;*/
 
@@ -73,9 +78,6 @@ rfbSendCursorShape(cl)
 	       sz_rfbFramebufferUpdateRectHeader);
 	cl->ublen += sz_rfbFramebufferUpdateRectHeader;
 
-	cl->cursorShapeBytesSent += sz_rfbFramebufferUpdateRectHeader;
-	cl->cursorShapeUpdatesSent++;
-
 	if (!rfbSendUpdateBuf(cl))
 	    return FALSE;
 
@@ -162,9 +164,8 @@ rfbSendCursorShape(cl)
     }
 
     /* Send everything we have prepared in the cl->updateBuf[]. */
-
-    cl->cursorShapeBytesSent += (cl->ublen - saved_ublen);
-    cl->cursorShapeUpdatesSent++;
+    rfbStatRecordEncodingSent(cl, (cl->useRichCursorEncoding ? rfbEncodingRichCursor : rfbEncodingXCursor), 
+        sz_rfbFramebufferUpdateRectHeader + (cl->ublen - saved_ublen), sz_rfbFramebufferUpdateRectHeader + (cl->ublen - saved_ublen));
 
     if (!rfbSendUpdateBuf(cl))
 	return FALSE;
@@ -196,8 +197,7 @@ rfbSendCursorPos(rfbClientPtr cl)
 	 sz_rfbFramebufferUpdateRectHeader);
   cl->ublen += sz_rfbFramebufferUpdateRectHeader;
 
-  cl->cursorPosBytesSent += sz_rfbFramebufferUpdateRectHeader;
-  cl->cursorPosUpdatesSent++;
+  rfbStatRecordEncodingSent(cl, rfbEncodingPointerPos, sz_rfbFramebufferUpdateRectHeader, sz_rfbFramebufferUpdateRectHeader);
 
   if (!rfbSendUpdateBuf(cl))
     return FALSE;
@@ -305,11 +305,51 @@ char* rfbMakeMaskForXCursor(int width,int height,char* source)
    return(mask);
 }
 
+/* this function dithers the alpha using Floyd-Steinberg */
+
+char* rfbMakeMaskFromAlphaSource(int width,int height,unsigned char* alphaSource)
+{
+	int* error=(int*)calloc(sizeof(int),width);
+	int i,j,currentError=0,maskStride=(width+7)/8;
+	unsigned char* result=(unsigned char*)calloc(maskStride,height);
+	
+	for(j=0;j<height;j++)
+		for(i=0;i<width;i++) {
+			int right,middle,left;
+			currentError+=alphaSource[i+width*j]+error[i];
+		
+			if(currentError<0x80) {
+				/* set to transparent */
+				/* alpha was treated as 0 */
+			} else {
+				/* set to solid */
+				result[i/8+j*maskStride]|=(0x100>>(i&7));
+				/* alpha was treated as 0xff */
+				currentError-=0xff;
+			}
+			/* propagate to next row */
+			right=currentError/16;
+			middle=currentError*5/16;
+			left=currentError*3/16;
+			currentError-=right+middle+left;
+			error[i]=right;
+			if(i>0) {
+				error[i-1]=middle;
+				if(i>1)
+					error[i-2]=left;
+			}
+		}
+	free(error);
+	return (char *) result;
+}
+
 void rfbFreeCursor(rfbCursorPtr cursor)
 {
    if(cursor) {
        if(cursor->cleanupRichSource && cursor->richSource)
 	   free(cursor->richSource);
+       if(cursor->cleanupRichSource && cursor->alphaSource)
+	   free(cursor->alphaSource);
        if(cursor->cleanupSource && cursor->source)
 	   free(cursor->source);
        if(cursor->cleanupMask && cursor->mask)
@@ -319,7 +359,8 @@ void rfbFreeCursor(rfbCursorPtr cursor)
        else {
 	   cursor->cleanup=cursor->cleanupSource=cursor->cleanupMask
 	       =cursor->cleanupRichSource=FALSE;
-	   cursor->source=cursor->mask=cursor->richSource=0;
+	   cursor->source=cursor->mask=cursor->richSource=NULL;
+	   cursor->alphaSource=NULL;
        }
    }
    
@@ -384,19 +425,20 @@ void rfbMakeRichCursorFromXCursor(rfbScreenInfoPtr rfbScreen,rfbCursorPtr cursor
 
 /* functions to draw/hide cursor directly in the frame buffer */
 
-void rfbUndrawCursor(rfbScreenInfoPtr s)
+void rfbHideCursor(rfbClientPtr cl)
 {
+   rfbScreenInfoPtr s=cl->screen;
    rfbCursorPtr c=s->cursor;
    int j,x1,x2,y1,y2,bpp=s->serverFormat.bitsPerPixel/8,
      rowstride=s->paddedWidthInBytes;
    LOCK(s->cursorMutex);
-   if(!s->cursorIsDrawn || !c) {
+   if(!c) {
      UNLOCK(s->cursorMutex);
      return;
    }
    
    /* restore what is under the cursor */
-   x1=s->cursorX-c->xhot;
+   x1=cl->cursorX-c->xhot;
    x2=x1+c->width;
    if(x1<0) x1=0;
    if(x2>=s->width) x2=s->width-1;
@@ -404,7 +446,7 @@ void rfbUndrawCursor(rfbScreenInfoPtr s)
      UNLOCK(s->cursorMutex);
      return;
    }
-   y1=s->cursorY-c->yhot;
+   y1=cl->cursorY-c->yhot;
    y2=y1+c->height;
    if(y1<0) y1=0;
    if(y2>=s->height) y2=s->height-1;
@@ -418,16 +460,16 @@ void rfbUndrawCursor(rfbScreenInfoPtr s)
      memcpy(s->frameBuffer+(y1+j)*rowstride+x1*bpp,
 	    s->underCursorBuffer+j*x2*bpp,
 	    x2*bpp);
+
+   /* Copy to all scaled versions */
+   rfbScaledScreenUpdate(s, x1, y1, x1+x2, y1+y2);
    
-   /* rfbMarkRectAsModified(s,x1,y1,x1+x2,y1+y2); */
-   s->cursorIsDrawn = FALSE;
-   s->oldCursorX=s->cursorX;
-   s->oldCursorY=s->cursorY;
    UNLOCK(s->cursorMutex);
 }
 
-void rfbDrawCursor(rfbScreenInfoPtr s)
+void rfbShowCursor(rfbClientPtr cl)
 {
+   rfbScreenInfoPtr s=cl->screen;
    rfbCursorPtr c=s->cursor;
    int i,j,x1,x2,y1,y2,i1,j1,bpp=s->serverFormat.bitsPerPixel/8,
      rowstride=s->paddedWidthInBytes,
@@ -436,18 +478,7 @@ void rfbDrawCursor(rfbScreenInfoPtr s)
 
    if(!c) return;
    LOCK(s->cursorMutex);
-   if(s->cursorIsDrawn) {
-     /* is already drawn */
-     UNLOCK(s->cursorMutex);
-     return;
-   }
 
-   if(s->cursor && s->underCursorBuffer &&
-		   (s->cursorX!=s->oldCursorX || s->cursorY!=s->oldCursorY)) {
-	   int x1=s->oldCursorX-s->cursor->xhot,x2=x1+s->cursor->width;
-	   int y1=s->oldCursorY-s->cursor->yhot,y2=y1+s->cursor->height;
-	   rfbMarkRectAsModified(s,x1,y1,x2,y2);
-   }
    bufSize=c->width*c->height*bpp;
    w=(c->width+7)/8;
    if(s->underCursorBufferLen<bufSize) {
@@ -456,9 +487,10 @@ void rfbDrawCursor(rfbScreenInfoPtr s)
       s->underCursorBuffer=malloc(bufSize);
       s->underCursorBufferLen=bufSize;
    }
+
    /* save what is under the cursor */
    i1=j1=0; /* offset in cursor */
-   x1=s->cursorX-c->xhot;
+   x1=cl->cursorX-c->xhot;
    x2=x1+c->width;
    if(x1<0) { i1=-x1; x1=0; }
    if(x2>=s->width) x2=s->width-1;
@@ -466,7 +498,8 @@ void rfbDrawCursor(rfbScreenInfoPtr s)
      UNLOCK(s->cursorMutex);
      return; /* nothing to do */
    }
-   y1=s->cursorY-c->yhot;
+
+   y1=cl->cursorY-c->yhot;
    y2=y1+c->height;
    if(y1<0) { j1=-y1; y1=0; }
    if(y2>=s->height) y2=s->height-1;
@@ -488,23 +521,146 @@ void rfbDrawCursor(rfbScreenInfoPtr s)
    
    if(!c->richSource)
      rfbMakeRichCursorFromXCursor(s,c);
-   
-   /* now the cursor has to be drawn */
-   for(j=0;j<y2;j++)
-     for(i=0;i<x2;i++)
-       if((c->mask[(j+j1)*w+(i+i1)/8]<<((i+i1)&7))&0x80)
-	 memcpy(s->frameBuffer+(j+y1)*rowstride+(i+x1)*bpp,
-		c->richSource+(j+j1)*c->width*bpp+(i+i1)*bpp,bpp);
+  
+   if (c->alphaSource) {
+	int rmax, rshift;
+	int gmax, gshift;
+	int bmax, bshift;
+	int amax = 255;	/* alphaSource is always 8bits of info per pixel */
+	unsigned int rmask, gmask, bmask;
+
+	rmax   = s->serverFormat.redMax;
+	gmax   = s->serverFormat.greenMax;
+	bmax   = s->serverFormat.blueMax;
+	rshift = s->serverFormat.redShift;
+	gshift = s->serverFormat.greenShift;
+	bshift = s->serverFormat.blueShift;
+
+	rmask = (rmax << rshift);
+	gmask = (gmax << gshift);
+	bmask = (bmax << bshift);
+
+	for(j=0;j<y2;j++) {
+		for(i=0;i<x2;i++) {
+			/*
+			 * we loop over the whole cursor ignoring c->mask[],
+			 * using the extracted alpha value instead.
+			 */
+			char *dest;
+			unsigned char *src, *aptr;
+			unsigned int val, dval, sval;
+			int rdst, gdst, bdst;		/* fb RGB */
+			int asrc, rsrc, gsrc, bsrc;	/* rich source ARGB */
+
+			dest = s->frameBuffer + (j+y1)*rowstride + (i+x1)*bpp;
+			src  = c->richSource  + (j+j1)*c->width*bpp + (i+i1)*bpp;
+			aptr = c->alphaSource + (j+j1)*c->width + (i+i1);
+
+			asrc = *aptr;
+			if (!asrc) {
+				continue;
+			}
+
+			if (bpp == 1) {
+				dval = *((unsigned char*) dest);
+				sval = *((unsigned char*) src);
+			} else if (bpp == 2) {
+				dval = *((unsigned short*) dest);
+				sval = *((unsigned short*) src);
+			} else if (bpp == 3) {
+				unsigned char *dst = (unsigned char *) dest;
+				dval = 0;
+				dval |= ((*(dst+0)) << 0);
+				dval |= ((*(dst+1)) << 8);
+				dval |= ((*(dst+2)) << 16);
+				sval = 0;
+				sval |= ((*(src+0)) << 0);
+				sval |= ((*(src+1)) << 8);
+				sval |= ((*(src+2)) << 16);
+			} else if (bpp == 4) {
+				dval = *((unsigned int*) dest);
+				sval = *((unsigned int*) src);
+			} else {
+				continue;
+			}
+
+			/* extract dest and src RGB */
+			rdst = (dval & rmask) >> rshift;	/* fb */
+			gdst = (dval & gmask) >> gshift;
+			bdst = (dval & bmask) >> bshift;
+
+			rsrc = (sval & rmask) >> rshift;	/* richcursor */
+			gsrc = (sval & gmask) >> gshift;
+			bsrc = (sval & bmask) >> bshift;
+
+			/* blend in fb data. */
+			if (! c->alphaPreMultiplied) {
+				rsrc = (asrc * rsrc)/amax;
+				gsrc = (asrc * gsrc)/amax;
+				bsrc = (asrc * bsrc)/amax;
+			}
+			rdst = rsrc + ((amax - asrc) * rdst)/amax;
+			gdst = gsrc + ((amax - asrc) * gdst)/amax;
+			bdst = bsrc + ((amax - asrc) * bdst)/amax;
+
+			val = 0;
+			val |= (rdst << rshift);
+			val |= (gdst << gshift);
+			val |= (bdst << bshift);
+
+			/* insert the cooked pixel into the fb */
+			memcpy(dest, &val, bpp);
+		}
+	}
+   } else {
+      /* now the cursor has to be drawn */
+      for(j=0;j<y2;j++)
+        for(i=0;i<x2;i++)
+          if((c->mask[(j+j1)*w+(i+i1)/8]<<((i+i1)&7))&0x80)
+   	 memcpy(s->frameBuffer+(j+y1)*rowstride+(i+x1)*bpp,
+   		c->richSource+(j+j1)*c->width*bpp+(i+i1)*bpp,bpp);
+   }
+
+   /* Copy to all scaled versions */
+   rfbScaledScreenUpdate(s, x1, y1, x1+x2, y1+y2);
 
-   if(wasChanged)
-     rfbMarkRectAsModified(s,x1,y1,x1+x2,y1+y2);
-   s->cursorIsDrawn = TRUE;
    UNLOCK(s->cursorMutex);
 }
 
-/* for debugging */
+/* 
+ * If enableCursorShapeUpdates is FALSE, and the cursor is hidden, make sure
+ * that if the frameBuffer was transmitted with a cursor drawn, then that
+ * region gets redrawn.
+ */
 
-void rfbPrintXCursor(rfbCursorPtr cursor)
+void rfbRedrawAfterHideCursor(rfbClientPtr cl,sraRegionPtr updateRegion)
+{
+    rfbScreenInfoPtr s = cl->screen;
+    rfbCursorPtr c = s->cursor;
+    
+    if(c) {
+	int x,y,x2,y2;
+
+	x = cl->cursorX-c->xhot;
+	y = cl->cursorY-c->yhot;
+	x2 = x+c->width;
+	y2 = y+c->height;
+
+	if(sraClipRect2(&x,&y,&x2,&y2,0,0,s->width,s->height)) {
+	    sraRegionPtr rect;
+	    rect = sraRgnCreateRect(x,y,x2,y2);
+	    if(updateRegion)
+	    	sraRgnOr(updateRegion,rect);
+	    else
+		    sraRgnOr(cl->modifiedRegion,rect);
+	    sraRgnDestroy(rect);
+	}
+    }
+}
+
+#ifdef DEBUG
+
+static void rfbPrintXCursor(rfbCursorPtr cursor)
 {
    int i,i1,j,w=(cursor->width+7)/8;
    unsigned char bit;
@@ -518,23 +674,36 @@ void rfbPrintXCursor(rfbCursorPtr cursor)
    }
 }
 
-void rfbSetCursor(rfbScreenInfoPtr rfbScreen,rfbCursorPtr c,rfbBool freeOld)
+#endif
+
+void rfbSetCursor(rfbScreenInfoPtr rfbScreen,rfbCursorPtr c)
 {
+  rfbClientIteratorPtr iterator;
+  rfbClientPtr cl;
+
   LOCK(rfbScreen->cursorMutex);
-  while(rfbScreen->cursorIsDrawn) {
-    UNLOCK(rfbScreen->cursorMutex);
-    rfbUndrawCursor(rfbScreen);
-    LOCK(rfbScreen->cursorMutex);
+
+  if(rfbScreen->cursor) {
+    iterator=rfbGetClientIterator(rfbScreen);
+    while((cl=rfbClientIteratorNext(iterator)))
+	if(!cl->enableCursorShapeUpdates)
+	  rfbRedrawAfterHideCursor(cl,NULL);
+    rfbReleaseClientIterator(iterator);
+
+    if(rfbScreen->cursor->cleanup)
+	 rfbFreeCursor(rfbScreen->cursor);
   }
 
-  free(rfbScreen->underCursorBuffer);
-  rfbScreen->underCursorBuffer=0;
-  rfbScreen->underCursorBufferLen=0;
-
-  if(rfbScreen->cursor && (freeOld || rfbScreen->cursor->cleanup))
-    rfbFreeCursor(rfbScreen->cursor);
-
   rfbScreen->cursor = c;
 
+  iterator=rfbGetClientIterator(rfbScreen);
+  while((cl=rfbClientIteratorNext(iterator))) {
+    cl->cursorWasChanged = TRUE;
+    if(!cl->enableCursorShapeUpdates)
+      rfbRedrawAfterHideCursor(cl,NULL);
+  }
+  rfbReleaseClientIterator(iterator);
+
   UNLOCK(rfbScreen->cursorMutex);
 }
+
