@@ -10,8 +10,12 @@
  *  see GPL (latest version) for full details
  */
 
+#ifdef __STRICT_ANSI__
+#define _BSD_SOURCE
+#endif
 #include <rfb/rfb.h>
 #include <rfb/rfbregion.h>
+#include "private.h"
 
 #include <stdarg.h>
 #include <errno.h>
@@ -34,21 +38,181 @@
 #include <signal.h>
 #include <time.h>
 
+static int extMutex_initialized = 0;
+static int logMutex_initialized = 0;
 #ifdef LIBVNCSERVER_HAVE_LIBPTHREAD
-MUTEX(logMutex);
+static MUTEX(logMutex);
+static MUTEX(extMutex);
 #endif
 
-int rfbEnableLogging=1;
+static int rfbEnableLogging=1;
 
 #ifdef LIBVNCSERVER_WORDS_BIGENDIAN
-char rfbEndianTest = 0;
+char rfbEndianTest = (1==0);
 #else
-char rfbEndianTest = -1;
+char rfbEndianTest = (1==1);
 #endif
 
-/* from rfbserver.c */
-void rfbIncrClientRef(rfbClientPtr cl);
-void rfbDecrClientRef(rfbClientPtr cl);
+/*
+ * Protocol extensions
+ */
+
+static rfbProtocolExtension* rfbExtensionHead = NULL;
+
+/*
+ * This method registers a list of new extensions.  
+ * It avoids same extension getting registered multiple times. 
+ * The order is not preserved if multiple extensions are
+ * registered at one-go.
+ */
+void
+rfbRegisterProtocolExtension(rfbProtocolExtension* extension)
+{
+	rfbProtocolExtension *head = rfbExtensionHead, *next = NULL;
+
+	if(extension == NULL)
+		return;
+
+	next = extension->next;
+
+	if (! extMutex_initialized) {
+		INIT_MUTEX(extMutex);
+		extMutex_initialized = 1;
+	}
+
+	LOCK(extMutex);
+
+	while(head != NULL) {
+		if(head == extension) {
+			UNLOCK(extMutex);
+			rfbRegisterProtocolExtension(next);
+			return;
+		}
+
+		head = head->next;
+	}
+
+	extension->next = rfbExtensionHead;
+	rfbExtensionHead = extension;
+
+	UNLOCK(extMutex);
+	rfbRegisterProtocolExtension(next);
+}
+
+/*
+ * This method unregisters a list of extensions.  
+ * These extensions won't be available for any new
+ * client connection. 
+ */
+void
+rfbUnregisterProtocolExtension(rfbProtocolExtension* extension)
+{
+
+	rfbProtocolExtension *cur = NULL, *pre = NULL;
+
+	if(extension == NULL)
+		return;
+
+	if (! extMutex_initialized) {
+		INIT_MUTEX(extMutex);
+		extMutex_initialized = 1;
+	}
+
+	LOCK(extMutex);
+
+	if(rfbExtensionHead == extension) {
+		rfbExtensionHead = rfbExtensionHead->next;
+		UNLOCK(extMutex);
+		rfbUnregisterProtocolExtension(extension->next);
+		return;
+	}
+
+	cur = pre = rfbExtensionHead;
+
+	while(cur) {
+		if(cur == extension) {
+			pre->next = cur->next;
+			break;
+		}
+		pre = cur;
+		cur = cur->next;
+	}
+
+	UNLOCK(extMutex);
+
+	rfbUnregisterProtocolExtension(extension->next);
+}
+
+rfbProtocolExtension* rfbGetExtensionIterator()
+{
+	LOCK(extMutex);
+	return rfbExtensionHead;
+}
+
+void rfbReleaseExtensionIterator()
+{
+	UNLOCK(extMutex);
+}
+
+rfbBool rfbEnableExtension(rfbClientPtr cl, rfbProtocolExtension* extension,
+	void* data)
+{
+	rfbExtensionData* extData;
+
+	/* make sure extension is not yet enabled. */
+	for(extData = cl->extensions; extData; extData = extData->next)
+		if(extData->extension == extension)
+			return FALSE;
+
+	extData = calloc(sizeof(rfbExtensionData),1);
+	extData->extension = extension;
+	extData->data = data;
+	extData->next = cl->extensions;
+	cl->extensions = extData;
+
+	return TRUE;
+}
+
+rfbBool rfbDisableExtension(rfbClientPtr cl, rfbProtocolExtension* extension)
+{
+	rfbExtensionData* extData;
+	rfbExtensionData* prevData = NULL;
+
+	for(extData = cl->extensions; extData; extData = extData->next) {
+		if(extData->extension == extension) {
+			if(extData->data)
+				free(extData->data);
+			if(prevData == NULL)
+				cl->extensions = extData->next;
+			else
+				prevData->next = extData->next;
+			return TRUE;
+		}
+		prevData = extData;
+	}
+
+	return FALSE;
+}
+
+void* rfbGetExtensionClientData(rfbClientPtr cl, rfbProtocolExtension* extension)
+{
+    rfbExtensionData* data = cl->extensions;
+
+    while(data && data->extension != extension)
+	data = data->next;
+
+    if(data == NULL) {
+	rfbLog("Extension is not enabled !\n");
+	/* rfbCloseClient(cl); */
+	return NULL;
+    }
+
+    return data->data;
+}
+
+/*
+ * Logging
+ */
 
 void rfbLogEnable(int enabled) {
   rfbEnableLogging=enabled;
@@ -58,7 +222,7 @@ void rfbLogEnable(int enabled) {
  * rfbLog prints a time-stamped message to the log file (stderr).
  */
 
-void
+static void
 rfbDefaultLog(const char *format, ...)
 {
     va_list args;
@@ -68,6 +232,11 @@ rfbDefaultLog(const char *format, ...)
     if(!rfbEnableLogging)
       return;
 
+    if (! logMutex_initialized) {
+      INIT_MUTEX(logMutex);
+      logMutex_initialized = 1;
+    }
+
     LOCK(logMutex);
     va_start(args, format);
 
@@ -95,8 +264,6 @@ void rfbScheduleCopyRegion(rfbScreenInfoPtr rfbScreen,sraRegionPtr copyRegion,in
    rfbClientIteratorPtr iterator;
    rfbClientPtr cl;
 
-   rfbUndrawCursor(rfbScreen);
-  
    iterator=rfbGetClientIterator(rfbScreen);
    while((cl=rfbClientIteratorNext(iterator))) {
      LOCK(cl->updateMutex);
@@ -133,22 +300,44 @@ void rfbScheduleCopyRegion(rfbScreenInfoPtr rfbScreen,sraRegionPtr copyRegion,in
        sraRgnOr(cl->modifiedRegion,modifiedRegionBackup);
        sraRgnDestroy(modifiedRegionBackup);
 
-#if 0
-       /* TODO: is this needed? Or does it mess up deferring? */
-       /* while(!sraRgnEmpty(cl->copyRegion)) */ {
-#ifdef LIBVNCSERVER_HAVE_LIBPTHREAD
-	 if(!cl->screen->backgroundLoop)
-#endif
-	   {
-	     sraRegionPtr updateRegion = sraRgnCreateRgn(cl->modifiedRegion);
-	     sraRgnOr(updateRegion,cl->copyRegion);
-	     UNLOCK(cl->updateMutex);
-	     rfbSendFramebufferUpdate(cl,updateRegion);
-	     sraRgnDestroy(updateRegion);
-	     continue;
-	   }
+       if(!cl->enableCursorShapeUpdates) {
+          /*
+           * n.b. (dx, dy) is the vector pointing in the direction the
+           * copyrect displacement will take place.  copyRegion is the
+           * destination rectangle (say), not the source rectangle.
+           */
+          sraRegionPtr cursorRegion;
+          int x = cl->cursorX - cl->screen->cursor->xhot;
+          int y = cl->cursorY - cl->screen->cursor->yhot;
+          int w = cl->screen->cursor->width;
+          int h = cl->screen->cursor->height;
+
+          cursorRegion = sraRgnCreateRect(x, y, x + w, y + h);
+          sraRgnAnd(cursorRegion, cl->copyRegion);
+          if(!sraRgnEmpty(cursorRegion)) {
+             /*
+              * current cursor rect overlaps with the copy region *dest*,
+              * mark it as modified since we won't copy-rect stuff to it.
+              */
+             sraRgnOr(cl->modifiedRegion, cursorRegion);
+          }
+          sraRgnDestroy(cursorRegion);
+
+          cursorRegion = sraRgnCreateRect(x, y, x + w, y + h);
+          /* displace it to check for overlap with copy region source: */
+          sraRgnOffset(cursorRegion, dx, dy);
+          sraRgnAnd(cursorRegion, cl->copyRegion);
+          if(!sraRgnEmpty(cursorRegion)) {
+             /*
+              * current cursor rect overlaps with the copy region *source*,
+              * mark the *displaced* cursorRegion as modified since we
+              * won't copyrect stuff to it.
+              */
+             sraRgnOr(cl->modifiedRegion, cursorRegion);
+          }
+          sraRgnDestroy(cursorRegion);
        }
-#endif
+
      } else {
        sraRgnOr(cl->modifiedRegion,copyRegion);
      }
@@ -167,8 +356,6 @@ void rfbDoCopyRegion(rfbScreenInfoPtr screen,sraRegionPtr copyRegion,int dx,int
     rowstride=screen->paddedWidthInBytes;
    char *in,*out;
 
-   rfbUndrawCursor(screen);
-  
    /* copy it, really */
    i = sraRgnGetReverseIterator(copyRegion,dx<0,dy<0);
    while(sraRgnIteratorNext(i,&rect)) {
@@ -185,6 +372,7 @@ void rfbDoCopyRegion(rfbScreenInfoPtr screen,sraRegionPtr copyRegion,int dx,int
 	 memmove(out,in,widthInBytes);
      }
    }
+   sraRgnReleaseIterator(i);
   
    rfbScheduleCopyRegion(screen,copyRegion,dx,dy);
 }
@@ -193,12 +381,14 @@ void rfbDoCopyRect(rfbScreenInfoPtr screen,int x1,int y1,int x2,int y2,int dx,in
 {
   sraRegionPtr region = sraRgnCreateRect(x1,y1,x2,y2);
   rfbDoCopyRegion(screen,region,dx,dy);
+  sraRgnDestroy(region);
 }
 
 void rfbScheduleCopyRect(rfbScreenInfoPtr screen,int x1,int y1,int x2,int y2,int dx,int dy)
 {
   sraRegionPtr region = sraRgnCreateRect(x1,y1,x2,y2);
   rfbScheduleCopyRegion(screen,region,dx,dy);
+  sraRgnDestroy(region);
 }
 
 void rfbMarkRegionAsModified(rfbScreenInfoPtr screen,sraRegionPtr modRegion)
@@ -217,6 +407,7 @@ void rfbMarkRegionAsModified(rfbScreenInfoPtr screen,sraRegionPtr modRegion)
    rfbReleaseClientIterator(iterator);
 }
 
+void rfbScaledScreenUpdate(rfbScreenInfoPtr screen, int x1, int y1, int x2, int y2);
 void rfbMarkRectAsModified(rfbScreenInfoPtr screen,int x1,int y1,int x2,int y2)
 {
    sraRegionPtr region;
@@ -224,20 +415,25 @@ void rfbMarkRectAsModified(rfbScreenInfoPtr screen,int x1,int y1,int x2,int y2)
 
    if(x1>x2) { i=x1; x1=x2; x2=i; }
    if(x1<0) x1=0;
-   if(x2>=screen->width) x2=screen->width-1;
+   if(x2>screen->width) x2=screen->width;
    if(x1==x2) return;
    
    if(y1>y2) { i=y1; y1=y2; y2=i; }
    if(y1<0) y1=0;
-   if(y2>=screen->height) y2=screen->height-1;
+   if(y2>screen->height) y2=screen->height;
    if(y1==y2) return;
-   
+
+   /* update scaled copies for this rectangle */
+   rfbScaledScreenUpdate(screen,x1,y1,x2,y2);
+
    region = sraRgnCreateRect(x1,y1,x2,y2);
    rfbMarkRegionAsModified(screen,region);
    sraRgnDestroy(region);
 }
 
 #ifdef LIBVNCSERVER_HAVE_LIBPTHREAD
+#include <unistd.h>
+
 static void *
 clientOutput(void *data)
 {
@@ -287,6 +483,7 @@ clientOutput(void *data)
 	sraRgnDestroy(updateRegion);
     }
 
+    /* Not reached. */
     return NULL;
 }
 
@@ -298,7 +495,40 @@ clientInput(void *data)
     pthread_create(&output_thread, NULL, clientOutput, (void *)cl);
 
     while (1) {
-        rfbProcessClientMessage(cl);
+	fd_set rfds, wfds, efds;
+	struct timeval tv;
+	int n;
+
+	FD_ZERO(&rfds);
+	FD_SET(cl->sock, &rfds);
+	FD_ZERO(&efds);
+	FD_SET(cl->sock, &efds);
+
+	/* Are we transferring a file in the background? */
+	FD_ZERO(&wfds);
+	if ((cl->fileTransfer.fd!=-1) && (cl->fileTransfer.sending==1))
+	    FD_SET(cl->sock, &wfds);
+
+	tv.tv_sec = 60; /* 1 minute */
+	tv.tv_usec = 0;
+	n = select(cl->sock + 1, &rfds, &wfds, &efds, &tv);
+	if (n < 0) {
+	    rfbLogPerror("ReadExact: select");
+	    break;
+	}
+	if (n == 0) /* timeout */
+	{
+            rfbSendFileTransferChunk(cl);
+	    continue;
+        }
+        
+        /* We have some space on the transmit queue, send some data */
+        if (FD_ISSET(cl->sock, &wfds))
+            rfbSendFileTransferChunk(cl);
+
+        if (FD_ISSET(cl->sock, &rfds) || FD_ISSET(cl->sock, &efds))
+            rfbProcessClientMessage(cl);
+
         if (cl->sock == -1) {
             /* Client has disconnected. */
             break;
@@ -323,11 +553,12 @@ listenerRun(void *data)
     int client_fd;
     struct sockaddr_in peer;
     rfbClientPtr cl;
-    size_t len;
+    socklen_t len;
 
     len = sizeof(peer);
 
     /* TODO: this thread wont die by restarting the server */
+    /* TODO: HTTP is not handled */
     while ((client_fd = accept(screen->listenSock, 
                                (struct sockaddr*)&peer, &len)) >= 0) {
         cl = rfbNewClient(screen,client_fd);
@@ -372,23 +603,20 @@ rfbDefaultPtrAddEvent(int buttonMask, int x, int y, rfbClientPtr cl)
 {
   rfbClientIteratorPtr iterator;
   rfbClientPtr other_client;
+  rfbScreenInfoPtr s = cl->screen;
 
-  if (x != cl->screen->cursorX || y != cl->screen->cursorY) {
-    if (cl->screen->cursorIsDrawn)
-      rfbUndrawCursor(cl->screen);
-    LOCK(cl->screen->cursorMutex);
-    if (!cl->screen->cursorIsDrawn) {
-      cl->screen->cursorX = x;
-      cl->screen->cursorY = y;
-    }
-    UNLOCK(cl->screen->cursorMutex);
+  if (x != s->cursorX || y != s->cursorY) {
+    LOCK(s->cursorMutex);
+    s->cursorX = x;
+    s->cursorY = y;
+    UNLOCK(s->cursorMutex);
 
     /* The cursor was moved by this client, so don't send CursorPos. */
     if (cl->enableCursorPosUpdates)
       cl->cursorWasMoved = FALSE;
 
     /* But inform all remaining clients about this cursor movement. */
-    iterator = rfbGetClientIterator(cl->screen);
+    iterator = rfbGetClientIterator(s);
     while ((other_client = rfbClientIteratorNext(iterator)) != NULL) {
       if (other_client != cl && other_client->enableCursorPosUpdates) {
 	other_client->cursorWasMoved = TRUE;
@@ -398,7 +626,7 @@ rfbDefaultPtrAddEvent(int buttonMask, int x, int y, rfbClientPtr cl)
   }
 }
 
-void rfbDefaultSetXCutText(char* text, int len, rfbClientPtr cl)
+static void rfbDefaultSetXCutText(char* text, int len, rfbClientPtr cl)
 {
 }
 
@@ -413,7 +641,7 @@ static rfbCursor myCursor =
    8, 7, 3, 3,
    0, 0, 0,
    0xffff, 0xffff, 0xffff,
-   0
+   NULL
 };
 #else
 static rfbCursor myCursor = 
@@ -427,17 +655,17 @@ static rfbCursor myCursor =
    width: 8, height: 7, xhot: 3, yhot: 3,
    foreRed: 0, foreGreen: 0, foreBlue: 0,
    backRed: 0xffff, backGreen: 0xffff, backBlue: 0xffff,
-   richSource: 0
+   richSource: NULL
 };
 #endif
 
-rfbCursorPtr rfbDefaultGetCursorPtr(rfbClientPtr cl)
+static rfbCursorPtr rfbDefaultGetCursorPtr(rfbClientPtr cl)
 {
    return(cl->screen->cursor);
 }
 
 /* response is cl->authChallenge vncEncrypted with passwd */
-rfbBool rfbDefaultPasswordCheck(rfbClientPtr cl,const char* response,int len)
+static rfbBool rfbDefaultPasswordCheck(rfbClientPtr cl,const char* response,int len)
 {
   int i;
   char *passwd=rfbDecryptPasswdFromFile(cl->screen->authPasswdData);
@@ -493,7 +721,7 @@ void rfbDoNothingWithClient(rfbClientPtr cl)
 {
 }
 
-enum rfbNewClientAction rfbDefaultNewClientHook(rfbClientPtr cl)
+static enum rfbNewClientAction rfbDefaultNewClientHook(rfbClientPtr cl)
 {
 	return RFB_CLIENT_ACCEPT;
 }
@@ -548,17 +776,22 @@ rfbScreenInfoPtr rfbGetScreen(int* argc,char** argv,
  int width,int height,int bitsPerSample,int samplesPerPixel,
  int bytesPerPixel)
 {
-   rfbScreenInfoPtr screen=malloc(sizeof(rfbScreenInfo));
+   rfbScreenInfoPtr screen=calloc(sizeof(rfbScreenInfo),1);
+
+   if (! logMutex_initialized) {
+     INIT_MUTEX(logMutex);
+     logMutex_initialized = 1;
+   }
 
-   INIT_MUTEX(logMutex);
 
    if(width&3)
      rfbErr("WARNING: Width (%d) is not a multiple of 4. VncViewer has problems with that.\n",width);
 
    screen->autoPort=FALSE;
-   screen->clientHead=0;
+   screen->clientHead=NULL;
+   screen->pointerClient=NULL;
    screen->port=5900;
-   screen->socketInitDone=FALSE;
+   screen->socketState=RFB_SOCKET_INIT;
 
    screen->inetdInitDone = FALSE;
    screen->inetdSock=-1;
@@ -566,7 +799,7 @@ rfbScreenInfoPtr rfbGetScreen(int* argc,char** argv,
    screen->udpSock=-1;
    screen->udpSockConnected=FALSE;
    screen->udpPort=0;
-   screen->udpClient=0;
+   screen->udpClient=NULL;
 
    screen->maxFd=0;
    screen->listenSock=-1;
@@ -582,7 +815,7 @@ rfbScreenInfoPtr rfbGetScreen(int* argc,char** argv,
    screen->alwaysShared = FALSE;
    screen->neverShared = FALSE;
    screen->dontDisconnect = FALSE;
-   screen->authPasswdData = 0;
+   screen->authPasswdData = NULL;
    screen->authPasswdFirstViewOnly = 1;
    
    screen->width = width;
@@ -596,9 +829,21 @@ rfbScreenInfoPtr rfbGetScreen(int* argc,char** argv,
    /* disable progressive updating per default */
    screen->progressiveSliceHeight = 0;
 
+   screen->listenInterface = htonl(INADDR_ANY);
+
+   screen->deferUpdateTime=5;
+   screen->maxRectsPerUpdate=50;
+
+   screen->handleEventsEagerly = FALSE;
+
+   screen->protocolMajorVersion = rfbProtocolMajorVersion;
+   screen->protocolMinorVersion = rfbProtocolMinorVersion;
+
+   screen->permitFileTransfer = FALSE;
+
    if(!rfbProcessArguments(screen,argc,argv)) {
      free(screen);
-     return 0;
+     return NULL;
    }
 
 #ifdef WIN32
@@ -618,8 +863,6 @@ rfbScreenInfoPtr rfbGetScreen(int* argc,char** argv,
 
    /* cursor */
 
-   screen->cursorIsDrawn = FALSE;
-   screen->dontSendFramebufferUpdate = FALSE;
    screen->cursorX=screen->cursorY=screen->underCursorBufferLen=0;
    screen->underCursorBuffer=NULL;
    screen->dontConvertRichCursorToXCursor = FALSE;
@@ -628,9 +871,6 @@ rfbScreenInfoPtr rfbGetScreen(int* argc,char** argv,
 
    IF_PTHREADS(screen->backgroundLoop = FALSE);
 
-   screen->deferUpdateTime=5;
-   screen->maxRectsPerUpdate=50;
-
    /* proc's and hook's */
 
    screen->kbdAddEvent = rfbDefaultKbdAddEvent;
@@ -640,7 +880,8 @@ rfbScreenInfoPtr rfbGetScreen(int* argc,char** argv,
    screen->getCursorPtr = rfbDefaultGetCursorPtr;
    screen->setTranslateFunction = rfbSetTranslateFunction;
    screen->newClientHook = rfbDefaultNewClientHook;
-   screen->displayHook = 0;
+   screen->displayHook = NULL;
+   screen->getKeyboardLedStateHook = NULL;
 
    /* initialize client list and iterator mutex */
    rfbClientListInit(screen);
@@ -666,10 +907,6 @@ void rfbNewFramebuffer(rfbScreenInfoPtr screen, char *framebuffer,
   rfbClientIteratorPtr iterator;
   rfbClientPtr cl;
 
-  /* Remove the pointer */
-
-  rfbUndrawCursor(screen);
-
   /* Update information in the screenInfo structure */
 
   old_format = screen->serverFormat;
@@ -726,9 +963,7 @@ void rfbNewFramebuffer(rfbScreenInfoPtr screen, char *framebuffer,
   rfbReleaseClientIterator(iterator);
 }
 
-#ifdef LIBVNCSERVER_HAVE_LIBJPEG
-extern void TightCleanup();
-#endif
+/* hang up on all clients and free all reserved memory */
 
 void rfbScreenCleanup(rfbScreenInfoPtr screen)
 {
@@ -741,24 +976,41 @@ void rfbScreenCleanup(rfbScreenInfoPtr screen)
   }
   rfbReleaseClientIterator(i);
     
-  /* TODO: hang up on all clients and free all reserved memory */
 #define FREE_IF(x) if(screen->x) free(screen->x)
   FREE_IF(colourMap.data.bytes);
   FREE_IF(underCursorBuffer);
   TINI_MUTEX(screen->cursorMutex);
-  if(screen->cursor)
+  if(screen->cursor && screen->cursor->cleanup)
     rfbFreeCursor(screen->cursor);
-  free(screen);
+
+  rfbRRECleanup(screen);
+  rfbCoRRECleanup(screen);
+  rfbUltraCleanup(screen);
+#ifdef LIBVNCSERVER_HAVE_LIBZ
+  rfbZlibCleanup(screen);
 #ifdef LIBVNCSERVER_HAVE_LIBJPEG
-  rfbTightCleanup();
+  rfbTightCleanup(screen);
 #endif
+
+  /* free all 'scaled' versions of this screen */
+  while (screen->scaledScreenNext!=NULL)
+  {
+      rfbScreenInfoPtr ptr;
+      ptr = screen->scaledScreenNext;
+      screen->scaledScreenNext = ptr->scaledScreenNext;
+      free(ptr->frameBuffer);
+      free(ptr);
+  }
+
+#endif
+  free(screen);
 }
 
 void rfbInitServer(rfbScreenInfoPtr screen)
 {
 #ifdef WIN32
   WSADATA trash;
-  int i=WSAStartup(MAKEWORD(2,2),&trash);
+  WSAStartup(MAKEWORD(2,2),&trash);
 #endif
   rfbInitSockets(screen);
   rfbHttpInitSockets(screen);
@@ -768,6 +1020,21 @@ void rfbInitServer(rfbScreenInfoPtr screen)
 #endif
 }
 
+void rfbShutdownServer(rfbScreenInfoPtr screen,rfbBool disconnectClients) {
+  if(disconnectClients) {
+    rfbClientPtr cl;
+    rfbClientIteratorPtr iter = rfbGetClientIterator(screen);
+    while( (cl = rfbClientIteratorNext(iter)) )
+      if (cl->sock > -1)
+	/* we don't care about maxfd here, because the server goes away */
+	rfbCloseClient(cl);
+    rfbReleaseClientIterator(iter);
+  }
+
+  rfbShutdownSockets(screen);
+  rfbHttpShutdownSockets(screen);
+}
+
 #ifndef LIBVNCSERVER_HAVE_GETTIMEOFDAY
 #include <fcntl.h>
 #include <conio.h>
@@ -782,15 +1049,15 @@ void gettimeofday(struct timeval* tv,char* dummy)
 }
 #endif
 
-/* defined in rfbserver.c, but kind of "private" */
-rfbClientPtr rfbClientIteratorHead(rfbClientIteratorPtr i);
-
-void
+rfbBool
 rfbProcessEvents(rfbScreenInfoPtr screen,long usec)
 {
   rfbClientIteratorPtr i;
   rfbClientPtr cl,clPrev;
   struct timeval tv;
+  rfbBool result=FALSE;
+  extern rfbClientIteratorPtr
+    rfbGetClientIteratorWithClosed(rfbScreenInfoPtr rfbScreen);
 
   if(usec<0)
     usec=screen->deferUpdateTime*1000;
@@ -801,11 +1068,12 @@ rfbProcessEvents(rfbScreenInfoPtr screen,long usec)
   corbaCheckFds(screen);
 #endif
 
-  i = rfbGetClientIterator(screen);
+  i = rfbGetClientIteratorWithClosed(screen);
   cl=rfbClientIteratorHead(i);
   while(cl) {
     if (cl->sock >= 0 && !cl->onHold && FB_UPDATE_PENDING(cl) &&
         !sraRgnEmpty(cl->requestedRegion)) {
+      result=TRUE;
       if(screen->deferUpdateTime == 0) {
 	  rfbSendFramebufferUpdate(cl,cl->modifiedRegion);
       } else if(cl->startDeferring.tv_usec == 0) {
@@ -823,12 +1091,41 @@ rfbProcessEvents(rfbScreenInfoPtr screen,long usec)
 	}
       }
     }
+
+    if (!cl->viewOnly && cl->lastPtrX >= 0) {
+      if(cl->startPtrDeferring.tv_usec == 0) {
+        gettimeofday(&cl->startPtrDeferring,NULL);
+        if(cl->startPtrDeferring.tv_usec == 0)
+          cl->startPtrDeferring.tv_usec++;
+      } else {
+        struct timeval tv;
+        gettimeofday(&tv,NULL);
+        if(tv.tv_sec < cl->startPtrDeferring.tv_sec /* at midnight */
+           || ((tv.tv_sec-cl->startPtrDeferring.tv_sec)*1000
+           +(tv.tv_usec-cl->startPtrDeferring.tv_usec)/1000)
+           > cl->screen->deferPtrUpdateTime) {
+          cl->startPtrDeferring.tv_usec = 0;
+          cl->screen->ptrAddEvent(cl->lastPtrButtons, 
+                                  cl->lastPtrX, 
+                                  cl->lastPtrY, cl);
+	  cl->lastPtrX = -1;
+        }
+      }
+    }
     clPrev=cl;
     cl=rfbClientIteratorNext(i);
-    if(clPrev->sock==-1)
+    if(clPrev->sock==-1) {
       rfbClientConnectionGone(clPrev);
+      result=TRUE;
+    }
   }
   rfbReleaseClientIterator(i);
+
+  return result;
+}
+
+rfbBool rfbIsActive(rfbScreenInfoPtr screenInfo) {
+  return screenInfo->socketState!=RFB_SOCKET_SHUTDOWN || screenInfo->clientHead!=NULL;
 }
 
 void rfbRunEventLoop(rfbScreenInfoPtr screen, long usec, rfbBool runInBackground)
@@ -850,6 +1147,6 @@ void rfbRunEventLoop(rfbScreenInfoPtr screen, long usec, rfbBool runInBackground
   if(usec<0)
     usec=screen->deferUpdateTime*1000;
 
-  while(1)
+  while(rfbIsActive(screen))
     rfbProcessEvents(screen,usec);
 }
