@@ -6,6 +6,8 @@
  */
 
 /*
+ *  Copyright (C) 2005 Rohit Kumar <rokumar@novell.com>,
+ *                     Johannes E. Schindelin <johannes.schindelin@gmx.de>
  *  Copyright (C) 2002 RealVNC Ltd.
  *  OSXvnc Copyright (C) 2001 Dan McGuirk <mcguirk@incompleteness.net>.
  *  Original Xvnc code Copyright (C) 1999 AT&T Laboratories Cambridge.  
@@ -60,8 +62,10 @@ extern "C"
 #define TINI_COND(cond) (rfbLog("%s:%d TINI_COND(%s)\n",__FILE__,__LINE__,#cond), pthread_cond_destroy(&(cond)))
 #define IF_PTHREADS(x) x
 #else
+#if !NONETWORK
 #define LOCK(mutex) pthread_mutex_lock(&(mutex));
 #define UNLOCK(mutex) pthread_mutex_unlock(&(mutex));
+#endif
 #define MUTEX(mutex) pthread_mutex_t (mutex)
 #define INIT_MUTEX(mutex) pthread_mutex_init(&(mutex),NULL)
 #define TINI_MUTEX(mutex) pthread_mutex_destroy(&(mutex))
@@ -116,6 +120,12 @@ enum rfbNewClientAction {
 	RFB_CLIENT_REFUSE
 };
 
+enum rfbSocketState {
+	RFB_SOCKET_INIT,
+	RFB_SOCKET_READY,
+	RFB_SOCKET_SHUTDOWN
+};
+
 typedef void (*rfbKbdAddEventProcPtr) (rfbBool down, rfbKeySym keySym, struct _rfbClientRec* cl);
 typedef void (*rfbKbdReleaseAllKeysProcPtr) (struct _rfbClientRec* cl);
 typedef void (*rfbPtrAddEventProcPtr) (int buttonMask, int x, int y, struct _rfbClientRec* cl);
@@ -125,6 +135,23 @@ typedef rfbBool (*rfbSetTranslateFunctionProcPtr)(struct _rfbClientRec* cl);
 typedef rfbBool (*rfbPasswordCheckProcPtr)(struct _rfbClientRec* cl,const char* encryptedPassWord,int len);
 typedef enum rfbNewClientAction (*rfbNewClientHookPtr)(struct _rfbClientRec* cl);
 typedef void (*rfbDisplayHookPtr)(struct _rfbClientRec* cl);
+/* support the capability to view the caps/num/scroll states of the X server */
+typedef int  (*rfbGetKeyboardLedStateHookPtr)(struct _rfbScreenInfo* screen);
+/* If x==1 and y==1 then set the whole display
+ * else find the window underneath x and y and set the framebuffer to the dimensions
+ * of that window
+ */
+typedef void (*rfbSetSingleWindowProcPtr) (struct _rfbClientRec* cl, int x, int y);
+/* Status determines if the X11 server permits input from the local user 
+ * status==0 or 1
+ */
+typedef void (*rfbSetServerInputProcPtr) (struct _rfbClientRec* cl, int status);
+/* Permit the server to allow or deny filetransfers.   This is defaulted to deny
+ * It is called when a client initiates a connection to determine if it is permitted.
+ */
+typedef int  (*rfbFileTransferPermitted) (struct _rfbClientRec* cl);
+/* Handle the textchat messages */
+typedef void (*rfbSetTextChat) (struct _rfbClientRec* cl, int length, char *string);
 
 typedef struct {
   uint32_t count;
@@ -135,6 +162,51 @@ typedef struct {
   } data; /* there have to be count*3 entries */
 } rfbColourMap;
 
+/*
+ * Security handling (RFB protocol version 3.7)
+ */
+
+typedef struct _rfbSecurity {
+	uint8_t type;
+	void (*handler)(struct _rfbClientRec* cl);
+	struct _rfbSecurity* next;
+} rfbSecurityHandler;
+
+/*
+ * Protocol extension handling.
+ */
+
+typedef struct _rfbProtocolExtension {
+	/* returns FALSE if extension should be deactivated for client.
+	   if newClient == NULL, it is always deactivated. */
+	rfbBool (*newClient)(struct _rfbClientRec* client, void** data);
+	/* returns FALSE if extension should be deactivated for client.
+	   if init == NULL, it stays activated. */
+	rfbBool (*init)(struct _rfbClientRec* client, void* data);
+	/* if pseudoEncodings is not NULL, it contains a 0 terminated
+	   list of the pseudo encodings handled by this extension. */
+	int *pseudoEncodings;
+	/* returns TRUE if that pseudo encoding is handled by the extension.
+	   encodingNumber==0 means "reset encodings". */
+	rfbBool (*enablePseudoEncoding)(struct _rfbClientRec* client,
+			void** data, int encodingNumber);
+	/* returns TRUE if message was handled */
+	rfbBool (*handleMessage)(struct _rfbClientRec* client,
+				void* data,
+				const rfbClientToServerMsg* message);
+	void (*close)(struct _rfbClientRec* client, void* data);
+	void (*usage)(void);
+	/* processArguments returns the number of handled arguments */
+	int (*processArgument)(int argc, char *argv[]);
+	struct _rfbProtocolExtension* next;
+} rfbProtocolExtension;
+
+typedef struct _rfbExtensionData {
+	rfbProtocolExtension* extension;
+	void* data;
+	struct _rfbExtensionData* next;
+} rfbExtensionData;
+
 /*
  * Per-screen (framebuffer) structure.  There can be as many as you wish,
  * each serving different clients. However, you have to call
@@ -143,6 +215,10 @@ typedef struct {
 
 typedef struct _rfbScreenInfo
 {
+    /* this structure has children that are scaled versions of this screen */
+    struct _rfbScreenInfo *scaledScreenNext;
+    int scaledScreenRefCount;
+
     int width;
     int paddedWidthInBytes;
     int height;
@@ -159,47 +235,6 @@ typedef struct _rfbScreenInfo
      */
     void* screenData;
   
-    /* The following two members are used to minimise the amount of unnecessary
-       drawing caused by cursor movement.  Whenever any drawing affects the
-       part of the screen where the cursor is, the cursor is removed first and
-       then the drawing is done (this is what the sprite routines test for).
-       Afterwards, however, we do not replace the cursor, even when the cursor
-       is logically being moved across the screen.  We only draw the cursor
-       again just as we are about to send the client a framebuffer update.
-
-       We need to be careful when removing and drawing the cursor because of
-       their relationship with the normal drawing routines.  The drawing
-       routines can invoke the cursor routines, but also the cursor routines
-       themselves end up invoking drawing routines.
-
-       Removing the cursor (rfbUndrawCursor) is eventually achieved by
-       doing a CopyArea from a pixmap to the screen, where the pixmap contains
-       the saved contents of the screen under the cursor.  Before doing this,
-       however, we set cursorIsDrawn to FALSE.  Then, when CopyArea is called,
-       it sees that cursorIsDrawn is FALSE and so doesn't feel the need to
-       (recursively!) remove the cursor before doing it.
-
-       Putting up the cursor (rfbDrawCursor) involves a call to
-       PushPixels.  While this is happening, cursorIsDrawn must be FALSE so
-       that PushPixels doesn't think it has to remove the cursor first.
-       Obviously cursorIsDrawn is set to TRUE afterwards.
-
-       Another problem we face is that drawing routines sometimes cause a
-       framebuffer update to be sent to the RFB client.  When the RFB client is
-       already waiting for a framebuffer update and some drawing to the
-       framebuffer then happens, the drawing routine sees that the client is
-       ready, so it calls rfbSendFramebufferUpdate.  If the cursor is not drawn
-       at this stage, it must be put up, and so rfbSpriteRestoreCursor is
-       called.  However, if the original drawing routine was actually called
-       from within rfbSpriteRestoreCursor or rfbSpriteRemoveCursor we don't
-       want this to happen.  So both the cursor routines set
-       dontSendFramebufferUpdate to TRUE, and all the drawing routines check
-       this before calling rfbSendFramebufferUpdate. */
-
-    rfbBool cursorIsDrawn;		    /* TRUE if the cursor is currently drawn */
-    rfbBool dontSendFramebufferUpdate; /* TRUE while removing or drawing the
-				       cursor */
-   
     /* additions by libvncserver */
 
     rfbPixelFormat serverFormat;
@@ -218,7 +253,7 @@ typedef struct _rfbScreenInfo
     fd_set allFds;
 #endif
 
-    rfbBool socketInitDone;
+    enum rfbSocketState socketState;
     SOCKET inetdSock;
     rfbBool inetdInitDone;
 
@@ -256,9 +291,11 @@ typedef struct _rfbScreenInfo
     rfbBool neverShared;
     rfbBool dontDisconnect;
     struct _rfbClientRec* clientHead;
+    struct _rfbClientRec* pointerClient;  /* "Mutex" for pointer events */
+
 
     /* cursor */
-    int cursorX, cursorY,oldCursorX,oldCursorY,underCursorBufferLen;
+    int cursorX, cursorY,underCursorBufferLen;
     char* underCursorBuffer;
     rfbBool dontConvertRichCursorToXCursor;
     struct rfbCursor* cursor;
@@ -273,12 +310,19 @@ typedef struct _rfbScreenInfo
     rfbSetXCutTextProcPtr setXCutText;
     rfbGetCursorProcPtr getCursorPtr;
     rfbSetTranslateFunctionProcPtr setTranslateFunction;
-  
+    rfbSetSingleWindowProcPtr setSingleWindow;
+    rfbSetServerInputProcPtr  setServerInput;
+    rfbFileTransferPermitted  getFileTransferPermission;
+    rfbSetTextChat            setTextChat;
+    
     /* newClientHook is called just after a new client is created */
     rfbNewClientHookPtr newClientHook;
     /* displayHook is called just before a frame buffer update */
     rfbDisplayHookPtr displayHook;
 
+    /* These hooks are called to pass keyboard state back to the client */
+    rfbGetKeyboardLedStateHookPtr getKeyboardLedStateHook;
+
 #ifdef LIBVNCSERVER_HAVE_LIBPTHREAD
     MUTEX(cursorMutex);
     rfbBool backgroundLoop;
@@ -291,6 +335,22 @@ typedef struct _rfbScreenInfo
      * an update should be sent. This should make working on a slow
      * link more interactive. */
     int progressiveSliceHeight;
+
+    in_addr_t listenInterface;
+    int deferPtrUpdateTime;
+
+    /* handle as many input events as possible (default off) */
+    rfbBool handleEventsEagerly;
+
+    /* rfbEncodingServerIdentity */
+    char *versionString;
+
+    /* What does the server tell the new clients which version it supports */
+    int protocolMajorVersion;
+    int protocolMinorVersion;
+
+    /* command line authorization of file transfers */
+    rfbBool permitFileTransfer;
 } rfbScreenInfo, *rfbScreenInfoPtr;
 
 
@@ -316,10 +376,37 @@ typedef struct sraRegion* sraRegionPtr;
 
 typedef void (*ClientGoneHookPtr)(struct _rfbClientRec* cl);
 
+typedef struct _rfbFileTransferData {
+  int fd;
+  int compressionEnabled;
+  int fileSize;
+  int numPackets;
+  int receiving;
+  int sending;
+} rfbFileTransferData;
+
+
+typedef struct _rfbStatList {
+    uint32_t type;
+    uint32_t sentCount;
+    uint32_t bytesSent;
+    uint32_t bytesSentIfRaw;
+    uint32_t rcvdCount;
+    uint32_t bytesRcvd;
+    uint32_t bytesRcvdIfRaw;
+    struct _rfbStatList *Next;
+} rfbStatList;
+
 typedef struct _rfbClientRec {
   
     /* back pointer to the screen */
     rfbScreenInfoPtr screen;
+
+     /* points to a scaled version of the screen buffer in cl->scaledScreenList */
+     rfbScreenInfoPtr scaledScreen;
+     /* how did the client tell us it wanted the screen changed?  Ultra style or palm style? */
+     rfbBool PalmVNC;
+    
   
     /* private data. You should put any application client specific data
      * into a struct and let clientData point to it. Don't forget to
@@ -333,12 +420,17 @@ typedef struct _rfbClientRec {
     SOCKET sock;
     char *host;
 
+    /* RFB protocol minor version number */
+    int protocolMajorVersion;
+    int protocolMinorVersion;
+
 #ifdef LIBVNCSERVER_HAVE_LIBPTHREAD
     pthread_t client_thread;
 #endif
                                 /* Possible client states: */
     enum {
         RFB_PROTOCOL_VERSION,   /* establishing protocol version */
+	RFB_SECURITY_TYPE,      /* negotiating security (RFB v.3.7) */
         RFB_AUTHENTICATION,     /* authenticating */
         RFB_INITIALISATION,     /* sending initialisation messages */
         RFB_NORMAL              /* normal protocol messages */
@@ -398,6 +490,10 @@ typedef struct _rfbClientRec {
        into a single update. */
 
       struct timeval startDeferring;
+      struct timeval startPtrDeferring;
+      int lastPtrX;
+      int lastPtrY;
+      int lastPtrButtons;
 
     /* translateFn points to the translation function which is used to copy
        and translate a rectangle from the framebuffer to an output buffer. */
@@ -418,20 +514,11 @@ typedef struct _rfbClientRec {
     int ublen;
 
     /* statistics */
-
-    int bytesSent[MAX_ENCODINGS];
-    int rectanglesSent[MAX_ENCODINGS];
-    int lastRectMarkersSent;
-    int lastRectBytesSent;
-    int cursorShapeBytesSent;
-    int cursorShapeUpdatesSent;
-    int cursorPosBytesSent;
-    int cursorPosUpdatesSent;
-    int framebufferUpdateMessagesSent;
+    struct _rfbStatList *statEncList;
+    struct _rfbStatList *statMsgList;
     int rawBytesEquivalent;
-    int keyEventsRcvd;
-    int pointerEventsRcvd;
-
+    int bytesSent;
+        
 #ifdef LIBVNCSERVER_HAVE_LIBZ
     /* zlib encoding -- necessary compression state info per client */
 
@@ -449,20 +536,29 @@ typedef struct _rfbClientRec {
 #endif
 #endif
 
+    /* Ultra Encoding support */
+    rfbBool compStreamInitedLZO;
+    char *lzoWrkMem;
+
+    rfbFileTransferData fileTransfer;
+
+    int     lastKeyboardLedState;     /* keep track of last value so we can send *change* events */
+    rfbBool enableSupportedMessages;  /* client supports SupportedMessages encoding */
+    rfbBool enableSupportedEncodings; /* client supports SupportedEncodings encoding */
+    rfbBool enableServerIdentity;     /* client supports ServerIdentity encoding */
+    rfbBool enableKeyboardLedState;   /* client supports KeyboardState encoding */
     rfbBool enableLastRectEncoding;   /* client supports LastRect encoding */
     rfbBool enableCursorShapeUpdates; /* client supports cursor shape updates */
     rfbBool enableCursorPosUpdates;   /* client supports cursor position updates */
     rfbBool useRichCursorEncoding;    /* rfbEncodingRichCursor is preferred */
     rfbBool cursorWasChanged;         /* cursor shape update should be sent */
     rfbBool cursorWasMoved;           /* cursor position update should be sent */
+    int cursorX,cursorY;	      /* the coordinates of the cursor,
+					 if enableCursorShapeUpdates = FALSE */
 
     rfbBool useNewFBSize;             /* client supports NewFBSize encoding */
     rfbBool newFBSizePending;         /* framebuffer size was changed */
 
-#ifdef LIBVNCSERVER_BACKCHANNEL
-    rfbBool enableBackChannel;        /* custom channel for special clients */
-#endif
-
     struct _rfbClientRec *prev;
     struct _rfbClientRec *next;
 
@@ -488,6 +584,8 @@ typedef struct _rfbClientRec {
     /* if progressive updating is on, this variable holds the current
      * y coordinate of the progressive slice. */
     int progressiveSliceY;
+
+    rfbExtensionData* extensions;
 } rfbClientRec, *rfbClientPtr;
 
 /*
@@ -496,8 +594,10 @@ typedef struct _rfbClientRec {
  */
 
 #define FB_UPDATE_PENDING(cl)                                              \
-     ((!(cl)->enableCursorShapeUpdates && !(cl)->screen->cursorIsDrawn) || \
-     ((cl)->enableCursorShapeUpdates && (cl)->cursorWasChanged) ||         \
+     (((cl)->enableCursorShapeUpdates && (cl)->cursorWasChanged) ||        \
+     (((cl)->enableCursorShapeUpdates == FALSE &&                          \
+       ((cl)->cursorX != (cl)->screen->cursorX ||                          \
+	(cl)->cursorY != (cl)->screen->cursorY))) ||                       \
      ((cl)->useNewFBSize && (cl)->newFBSizePending) ||                     \
      ((cl)->enableCursorPosUpdates && (cl)->cursorWasMoved) ||             \
      !sraRgnEmpty((cl)->copyRegion) || !sraRgnEmpty((cl)->modifiedRegion))
@@ -523,21 +623,28 @@ extern char rfbEndianTest;
 #define Swap24IfLE(l) (rfbEndianTest ? Swap24(l) : (l))
 #define Swap32IfLE(l) (rfbEndianTest ? Swap32(l) : (l))
 
+/* UltraVNC uses some windows structures unmodified, so the viewer expects LittleEndian Data */
+#define Swap16IfBE(s) (rfbEndianTest ? (s) : Swap16(s))
+#define Swap24IfBE(l) (rfbEndianTest ? (l) : Swap24(l))
+#define Swap32IfBE(l) (rfbEndianTest ? (l) : Swap32(l))
+
 /* sockets.c */
 
 extern int rfbMaxClientWait;
 
 extern void rfbInitSockets(rfbScreenInfoPtr rfbScreen);
+extern void rfbShutdownSockets(rfbScreenInfoPtr rfbScreen);
 extern void rfbDisconnectUDPSock(rfbScreenInfoPtr rfbScreen);
 extern void rfbCloseClient(rfbClientPtr cl);
 extern int rfbReadExact(rfbClientPtr cl, char *buf, int len);
 extern int rfbReadExactTimeout(rfbClientPtr cl, char *buf, int len,int timeout);
 extern int rfbWriteExact(rfbClientPtr cl, const char *buf, int len);
-extern void rfbCheckFds(rfbScreenInfoPtr rfbScreen,long usec);
+extern int rfbCheckFds(rfbScreenInfoPtr rfbScreen,long usec);
 extern int rfbConnect(rfbScreenInfoPtr rfbScreen, char* host, int port);
 extern int rfbConnectToTcpAddr(char* host, int port);
-extern int rfbListenOnTCPPort(int port);
-extern int rfbListenOnUDPPort(int port);
+extern int rfbListenOnTCPPort(int port, in_addr_t iface);
+extern int rfbListenOnUDPPort(int port, in_addr_t iface);
+extern int rfbStringToAddr(char* string,in_addr_t* addr);
 
 /* rfbserver.c */
 
@@ -549,6 +656,8 @@ extern void rfbClientListInit(rfbScreenInfoPtr rfbScreen);
 extern rfbClientIteratorPtr rfbGetClientIterator(rfbScreenInfoPtr rfbScreen);
 extern rfbClientPtr rfbClientIteratorNext(rfbClientIteratorPtr iterator);
 extern void rfbReleaseClientIterator(rfbClientIteratorPtr iterator);
+extern void rfbIncrClientRef(rfbClientPtr cl);
+extern void rfbDecrClientRef(rfbClientPtr cl);
 
 extern void rfbNewClientConnection(rfbScreenInfoPtr rfbScreen,int sock);
 extern rfbClientPtr rfbNewClient(rfbScreenInfoPtr rfbScreen,int sock);
@@ -569,11 +678,14 @@ extern rfbBool rfbSendNewFBSize(rfbClientPtr cl, int w, int h);
 extern rfbBool rfbSendSetColourMapEntries(rfbClientPtr cl, int firstColour, int nColours);
 extern void rfbSendBell(rfbScreenInfoPtr rfbScreen);
 
-void rfbGotXCutText(rfbScreenInfoPtr rfbScreen, char *str, int len);
+extern char *rfbProcessFileTransferReadBuffer(rfbClientPtr cl, uint32_t length);
+extern rfbBool rfbSendFileTransferChunk(rfbClientPtr cl);
+extern rfbBool rfbSendDirContent(rfbClientPtr cl, int length, char *buffer);
+extern rfbBool rfbSendFileTransferMessage(rfbClientPtr cl, uint8_t contentType, uint8_t contentParam, uint32_t size, uint32_t length, char *buffer);
+extern char *rfbProcessFileTransferReadBuffer(rfbClientPtr cl, uint32_t length);
+extern rfbBool rfbProcessFileTransfer(rfbClientPtr cl, uint8_t contentType, uint8_t contentParam, uint32_t size, uint32_t length);
 
-#ifdef LIBVNCSERVER_BACKCHANNEL
-extern void rfbSendBackChannel(rfbScreenInfoPtr s,char* message,int len);
-#endif
+void rfbGotXCutText(rfbScreenInfoPtr rfbScreen, char *str, int len);
 
 /* translate.c */
 
@@ -591,6 +703,7 @@ extern void rfbSetClientColourMaps(rfbScreenInfoPtr rfbScreen, int firstColour,
 /* httpd.c */
 
 extern void rfbHttpInitSockets(rfbScreenInfoPtr rfbScreen);
+extern void rfbHttpShutdownSockets(rfbScreenInfoPtr rfbScreen);
 extern void rfbHttpCheckFds(rfbScreenInfoPtr rfbScreen);
 
 
@@ -598,8 +711,10 @@ extern void rfbHttpCheckFds(rfbScreenInfoPtr rfbScreen);
 /* auth.c */
 
 extern void rfbAuthNewClient(rfbClientPtr cl);
+extern void rfbProcessClientSecurityType(rfbClientPtr cl);
 extern void rfbAuthProcessClientMessage(rfbClientPtr cl);
-
+extern void rfbRegisterSecurityHandler(rfbSecurityHandler* handler);
+extern void rfbUnregisterSecurityHandler(rfbSecurityHandler* handler);
 
 /* rre.c */
 
@@ -616,6 +731,17 @@ extern rfbBool rfbSendRectEncodingCoRRE(rfbClientPtr cl, int x,int y,int w,int h
 extern rfbBool rfbSendRectEncodingHextile(rfbClientPtr cl, int x, int y, int w,
                                        int h);
 
+/* ultra.c */
+
+/* Set maximum ultra rectangle size in pixels.  Always allow at least
+ * two scan lines.
+ */
+#define ULTRA_MAX_RECT_SIZE (128*256)
+#define ULTRA_MAX_SIZE(min) ((( min * 2 ) > ULTRA_MAX_RECT_SIZE ) ? \
+                            ( min * 2 ) : ULTRA_MAX_RECT_SIZE )
+
+extern rfbBool rfbSendRectEncodingUltra(rfbClientPtr cl, int x,int y,int w,int h);
+
 
 #ifdef LIBVNCSERVER_HAVE_LIBZ
 /* zlib.c */
@@ -644,6 +770,7 @@ extern rfbBool rfbTightDisableGradient;
 
 extern int rfbNumCodedRectsTight(rfbClientPtr cl, int x,int y,int w,int h);
 extern rfbBool rfbSendRectEncodingTight(rfbClientPtr cl, int x,int y,int w,int h);
+
 #endif
 #endif
 
@@ -659,6 +786,8 @@ typedef struct rfbCursor {
     unsigned short foreRed, foreGreen, foreBlue; /* device-independent colour */
     unsigned short backRed, backGreen, backBlue; /* device-independent colour */
     unsigned char *richSource; /* source bytes for a rich cursor */
+    unsigned char *alphaSource; /* source for alpha blending info */
+    rfbBool alphaPreMultiplied; /* if richSource already has alpha applied */
 } rfbCursor, *rfbCursorPtr;
 extern unsigned char rfbReverseByte[0x100];
 
@@ -667,12 +796,11 @@ extern rfbBool rfbSendCursorPos(rfbClientPtr cl);
 extern void rfbConvertLSBCursorBitmapOrMask(int width,int height,unsigned char* bitmap);
 extern rfbCursorPtr rfbMakeXCursor(int width,int height,char* cursorString,char* maskString);
 extern char* rfbMakeMaskForXCursor(int width,int height,char* cursorString);
+extern char* rfbMakeMaskFromAlphaSource(int width,int height,unsigned char* alphaSource);
 extern void rfbMakeXCursorFromRichCursor(rfbScreenInfoPtr rfbScreen,rfbCursorPtr cursor);
 extern void rfbMakeRichCursorFromXCursor(rfbScreenInfoPtr rfbScreen,rfbCursorPtr cursor);
 extern void rfbFreeCursor(rfbCursorPtr cursor);
-extern void rfbDrawCursor(rfbScreenInfoPtr rfbScreen);
-extern void rfbUndrawCursor(rfbScreenInfoPtr rfbScreen);
-extern void rfbSetCursor(rfbScreenInfoPtr rfbScreen,rfbCursorPtr c,rfbBool freeOld);
+extern void rfbSetCursor(rfbScreenInfoPtr rfbScreen,rfbCursorPtr c);
 
 /* cursor handling for the pointer */
 extern void rfbDefaultPtrAddEvent(int buttonMask,int x,int y,rfbClientPtr cl);
@@ -680,7 +808,6 @@ extern void rfbDefaultPtrAddEvent(int buttonMask,int x,int y,rfbClientPtr cl);
 /* zrle.c */
 #ifdef LIBVNCSERVER_HAVE_LIBZ
 extern rfbBool rfbSendRectEncodingZRLE(rfbClientPtr cl, int x, int y, int w,int h);
-extern void rfbFreeZrleData(rfbClientPtr cl);
 #endif
 
 /* stats.c */
@@ -718,7 +845,6 @@ void rfbFreeFont(rfbFontDataPtr font);
 
 /* draw.c */
 
-/* You have to call rfbUndrawCursor before using these functions */
 void rfbFillRect(rfbScreenInfoPtr s,int x1,int y1,int x2,int y2,rfbPixel col);
 void rfbDrawPixel(rfbScreenInfoPtr s,int x,int y,rfbPixel col);
 void rfbDrawLine(rfbScreenInfoPtr s,int x1,int y1,int x2,int y2,rfbPixel col);
@@ -760,6 +886,14 @@ void rfbMarkRectAsModified(rfbScreenInfoPtr rfbScreen,int x1,int y1,int x2,int y
 void rfbMarkRegionAsModified(rfbScreenInfoPtr rfbScreen,sraRegionPtr modRegion);
 void rfbDoNothingWithClient(rfbClientPtr cl);
 enum rfbNewClientAction defaultNewClientHook(rfbClientPtr cl);
+void rfbRegisterProtocolExtension(rfbProtocolExtension* extension);
+void rfbUnregisterProtocolExtension(rfbProtocolExtension* extension);
+struct _rfbProtocolExtension* rfbGetExtensionIterator();
+void rfbReleaseExtensionIterator();
+rfbBool rfbEnableExtension(rfbClientPtr cl, rfbProtocolExtension* extension,
+	void* data);
+rfbBool rfbDisableExtension(rfbClientPtr cl, rfbProtocolExtension* extension);
+void* rfbGetExtensionClientData(rfbClientPtr cl, rfbProtocolExtension* extension);
 
 /* to check against plain passwords */
 rfbBool rfbCheckPasswordByList(rfbClientPtr cl,const char* response,int len);
@@ -769,11 +903,13 @@ extern rfbScreenInfoPtr rfbGetScreen(int* argc,char** argv,
  int width,int height,int bitsPerSample,int samplesPerPixel,
  int bytesPerPixel);
 extern void rfbInitServer(rfbScreenInfoPtr rfbScreen);
+extern void rfbShutdownServer(rfbScreenInfoPtr rfbScreen,rfbBool disconnectClients);
 extern void rfbNewFramebuffer(rfbScreenInfoPtr rfbScreen,char *framebuffer,
  int width,int height, int bitsPerSample,int samplesPerPixel,
  int bytesPerPixel);
 
 extern void rfbScreenCleanup(rfbScreenInfoPtr screenInfo);
+extern void rfbSetServerVersionIdentity(rfbScreenInfoPtr screen, char *fmt, ...);
 
 /* functions to accept/refuse a client that has been put on hold
    by a NewClientHookPtr function. Must not be called in other
@@ -784,10 +920,50 @@ extern void rfbRefuseOnHoldClient(rfbClientPtr cl);
 /* call one of these two functions to service the vnc clients.
  usec are the microseconds the select on the fds waits.
  if you are using the event loop, set this to some value > 0, so the
- server doesn't get a high load just by listening. */
+ server doesn't get a high load just by listening.
+ rfbProcessEvents() returns TRUE if an update was pending. */
 
 extern void rfbRunEventLoop(rfbScreenInfoPtr screenInfo, long usec, rfbBool runInBackground);
-extern void rfbProcessEvents(rfbScreenInfoPtr screenInfo,long usec);
+extern rfbBool rfbProcessEvents(rfbScreenInfoPtr screenInfo,long usec);
+extern rfbBool rfbIsActive(rfbScreenInfoPtr screenInfo);
+
+/* TightVNC file transfer extension */
+void rfbRegisterTightVNCFileTransferExtension();
+void rfbUnregisterTightVNCFileTransferExtension(); 
+
+/* Statistics */
+extern char *messageNameServer2Client(uint32_t type, char *buf, int len);
+extern char *messageNameClient2Server(uint32_t type, char *buf, int len);
+extern char *encodingName(uint32_t enc, char *buf, int len);
+
+extern rfbStatList *rfbStatLookupEncoding(rfbClientPtr cl, uint32_t type);
+extern rfbStatList *rfbStatLookupMessage(rfbClientPtr cl, uint32_t type);
+
+/* Each call to rfbStatRecord* adds one to the rect count for that type */
+extern void rfbStatRecordEncodingSent(rfbClientPtr cl, uint32_t type, int byteCount, int byteIfRaw);
+extern void rfbStatRecordEncodingSentAdd(rfbClientPtr cl, uint32_t type, int byteCount); /* Specifically for tight encoding */
+extern void rfbStatRecordEncodingRcvd(rfbClientPtr cl, uint32_t type, int byteCount, int byteIfRaw);
+extern void rfbStatRecordMessageSent(rfbClientPtr cl, uint32_t type, int byteCount, int byteIfRaw);
+extern void rfbStatRecordMessageRcvd(rfbClientPtr cl, uint32_t type, int byteCount, int byteIfRaw);
+extern void rfbResetStats(rfbClientPtr cl);
+extern void rfbPrintStats(rfbClientPtr cl);
+
+extern int rfbStatGetSentBytes(rfbClientPtr cl);
+extern int rfbStatGetSentBytesIfRaw(rfbClientPtr cl);
+extern int rfbStatGetRcvdBytes(rfbClientPtr cl);
+extern int rfbStatGetRcvdBytesIfRaw(rfbClientPtr cl);
+extern int rfbStatGetMessageCountSent(rfbClientPtr cl, uint32_t type);
+extern int rfbStatGetMessageCountRcvd(rfbClientPtr cl, uint32_t type);
+extern int rfbStatGetEncodingCountSent(rfbClientPtr cl, uint32_t type);
+extern int rfbStatGetEncodingCountRcvd(rfbClientPtr cl, uint32_t type);
+
+/* Set which version you want to advertise 3.3, 3.6, 3.7 and 3.8 are currently supported*/
+extern void rfbSetProtocolVersion(rfbScreenInfoPtr rfbScreen, int major_, int minor_);
+
+/* send a TextChat message to a client */
+extern rfbBool rfbSendTextChatMessage(rfbClientPtr cl, uint32_t length, char *buffer);
+
+
 
 #endif
 
