@@ -63,6 +63,15 @@ typedef struct {
   rfbBool doNotSleep;
 } rfbVNCRec;
 
+/* client data */
+
+typedef struct rfbClientData {
+	void* tag;
+	void* data;
+	struct rfbClientData* next;
+} rfbClientData;
+
+/* app data (belongs into rfbClient?) */
 
 typedef struct {
   rfbBool shareDesktop;
@@ -80,11 +89,15 @@ typedef struct {
   int qualityLevel;
   rfbBool enableJPEG;
   rfbBool useRemoteCursor;
+  rfbBool palmVNC;  /* use palmvnc specific SetScale (vs ultravnc) */
+  int scaleSetting; /* 0 means no scale set, else 1/scaleSetting */
 } AppData;
 
 
 struct _rfbClient;
 
+typedef void (*HandleTextChatProc)(struct _rfbClient* client, int value, char *text);
+typedef void (*HandleKeyboardLedStateProc)(struct _rfbClient* client, int value, int pad);
 typedef rfbBool (*HandleCursorPosProc)(struct _rfbClient* client, int x, int y);
 typedef void (*SoftCursorLockAreaProc)(struct _rfbClient* client, int x, int y, int w, int h);
 typedef void (*SoftCursorUnlockScreenProc)(struct _rfbClient* client);
@@ -132,16 +145,74 @@ typedef struct _rfbClient {
 	char *bufoutptr;
 	int buffered;
 
+	/* The zlib encoding requires expansion/decompression/deflation of the
+	   compressed data in the "buffer" above into another, result buffer.
+	   However, the size of the result buffer can be determined precisely
+	   based on the bitsPerPixel, height and width of the rectangle.  We
+	   allocate this buffer one time to be the full size of the buffer. */
+
+	/* Ultra Encoding uses this buffer too */
+	
+	int ultra_buffer_size;
+	char *ultra_buffer;
+
+	int raw_buffer_size;
+	char *raw_buffer;
+
+#ifdef LIBVNCSERVER_HAVE_LIBZ
+	z_stream decompStream;
+	rfbBool decompStreamInited;
+#endif
+
+
+#ifdef LIBVNCSERVER_HAVE_LIBZ
+#ifdef LIBVNCSERVER_HAVE_LIBJPEG
+	/*
+	 * Variables for the ``tight'' encoding implementation.
+	 */
+
+	/* Separate buffer for compressed data. */
+#define ZLIB_BUFFER_SIZE 30000
+	char zlib_buffer[ZLIB_BUFFER_SIZE];
+
+	/* Four independent compression streams for zlib library. */
+	z_stream zlibStream[4];
+	rfbBool zlibStreamActive[4];
+
+	/* Filter stuff. Should be initialized by filter initialization code. */
+	rfbBool cutZeros;
+	int rectWidth, rectColors;
+	char tightPalette[256*4];
+	uint8_t tightPrevRow[2048*3*sizeof(uint16_t)];
+
+	/* JPEG decoder state. */
+	rfbBool jpegError;
+
+	struct jpeg_source_mgr* jpegSrcManager;
+	void* jpegBufferPtr;
+	size_t jpegBufferLen;
+
+#endif
+#endif
+
 
 	/* cursor.c */
 	uint8_t *rcSource, *rcMask;
 
 	/* private data pointer */
-	void* clientData;
+	rfbClientData* clientData;
 
 	rfbVNCRec* vncRec;
 
+	/* Keyboard State support (is 'Caps Lock' set on the remote display???) */
+	int KeyboardLedStateEnabled;
+	int CurrentKeyboardLedState;
+
+	int canHandleNewFBSize;
+
 	/* hooks */
+	HandleTextChatProc         HandleTextChat;
+	HandleKeyboardLedStateProc HandleKeyboardLedState;
 	HandleCursorPosProc HandleCursorPos;
 	SoftCursorLockAreaProc SoftCursorLockArea;
 	SoftCursorUnlockScreenProc SoftCursorUnlockScreen;
@@ -150,6 +221,19 @@ typedef struct _rfbClient {
 	GetPasswordProc GetPassword;
 	MallocFrameBufferProc MallocFrameBuffer;
 	BellProc Bell;
+
+	/* Which messages are supported by the server
+	 * This is a *guess* for most servers.
+	 * (If we can even detect the type of server)
+	 *
+	 * If the server supports the "rfbEncodingSupportedMessages"
+	 * then this will be updated when the encoding is received to
+	 * accurately reflect the servers capabilities.
+	 */
+	rfbSupportedMessages supportedMessages;
+
+	/* negotiated protocol version */
+	int major, minor;
 } rfbClient;
 
 /* cursor.c */
@@ -172,13 +256,40 @@ extern rfbBool SendIncrementalFramebufferUpdateRequest(rfbClient* client);
 extern rfbBool SendFramebufferUpdateRequest(rfbClient* client,
 					 int x, int y, int w, int h,
 					 rfbBool incremental);
+extern rfbBool SendScaleSetting(rfbClient* client,int scaleSetting);
 extern rfbBool SendPointerEvent(rfbClient* client,int x, int y, int buttonMask);
 extern rfbBool SendKeyEvent(rfbClient* client,uint32_t key, rfbBool down);
 extern rfbBool SendClientCutText(rfbClient* client,char *str, int len);
 extern rfbBool HandleRFBServerMessage(rfbClient* client);
 
+extern rfbBool TextChatSend(rfbClient* client, char *text);
+extern rfbBool TextChatOpen(rfbClient* client);
+extern rfbBool TextChatClose(rfbClient* client);
+extern rfbBool TextChatFinish(rfbClient* client);
+extern rfbBool PermitServerInput(rfbClient* client, int enabled);
+
 extern void PrintPixelFormat(rfbPixelFormat *format);
 
+/* client data */
+
+void rfbClientSetClientData(rfbClient* client, void* tag, void* data);
+void* rfbClientGetClientData(rfbClient* client, void* tag);
+
+/* protocol extensions */
+
+typedef struct _rfbClientProtocolExtension {
+	int* encodings;
+	/* returns TRUE if the encoding was handled */
+	rfbBool (*handleEncoding)(rfbClient* cl,
+		rfbFramebufferUpdateRectHeader* rect);
+	/* returns TRUE if it handled the message */
+	rfbBool (*handleMessage)(rfbClient* cl,
+		 rfbServerToClientMsg* message);
+	struct _rfbClientProtocolExtension* next;
+} rfbClientProtocolExtension;
+
+void rfbClientRegisterExtension(rfbClientProtocolExtension* e);
+
 /* sockets.c */
 
 extern rfbBool errorMessageOnReadFailure;
@@ -198,6 +309,7 @@ extern int WaitForMessage(rfbClient* client,unsigned int usecs);
 /* vncviewer.c */
 rfbClient* rfbGetClient(int bitsPerSample,int samplesPerPixel,int bytesPerPixel);
 rfbBool rfbInitClient(rfbClient* client,int* argc,char** argv);
+/* rfbClientCleanup() does not touch client->frameBuffer */
 void rfbClientCleanup(rfbClient* client);
 
 #endif
